//************************************************************
//  University of Patras
//	Department of Electrical Engineering & Computer Technology
//
//	Course: Parallel Programming in AI
//	Week 04: Code Optimization
//
//  Description: Arithmetic operations of Large two dimensional Matrices
//  	addmatOrg06.c: the version addmatOrg02.c using the parallel
//						programming directives of OpenMP API
//	Author: Evangelos Dermatas
//	v0.1
/*
*********************************************************************
gcc -v:	gcc version 7.4.0 (Ubuntu 7.4.0-1ubuntu1~18.04.1)
OS: 	Linux Ubuntu 18.04
CPU:	8-core, Intel(R) Core(TM) i7-2630QM CPU @ 2.00GHz, 3991 BogoMips
*********************************************************************
Compiling:	gcc addmatOrg06.c -o addmatOrg06 -fopenmp
Executing:	time ./addmatOrg06
Output in my Laptop:
innerProdMat  60312134

real	0m2.549s
user	0m18.775s
sys		0m1.320s
*********************************************************************
Compiling:	gcc addmatOrg06.c -o addmatOrg06 -fopenmp -O2
Executing:	time ./addmatOrg06
Output in my Laptop:

innerProdMat  -239170325

real	0m2.588s
user	0m19.055s
sys		0m1.337s
*********************************************************************
Compiling:	gcc addmatOrg06.c -o addmatOrg06 -fopenmp -O4
Executing:	time ./addmatOrg06
Output in my Laptop:

innerProdMat  40394700

real	0m2.585s
user	0m18.993s
sys		0m1.315s
*********************************************************************
Compiling:	icc addmatOrg06.c -o addmatOrg06 -fopenmp
Executing:	time ./addmatOrg06
Output in my Laptop:

innerProdMat  -74668125

real	0m0.441s
user	0m1.958s
sys		0m1.288s

*/
//************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <omp.h>

// Matrices Dimension
#define N	1000

// **************************************************************
int A[N][N] ;
int B[N][N] ;
int C[N][N] ;

// ***************************************************************************
//    Set random values to the elements of a two dimensional matrix a[N][N]
// 		in the range [-50,49]
// ***************************************************************************
void setValMat( int a[N][N] ) 
{
	int i, j ;

//#pragma omp parallel for
	for( i = 0 ; i < N ; i++ )
//#pragma omp simd
	for( j = 0 ; j < N ; j++ )
		    a[i][j] = (rand() % 100) - 50 ;

}


// ***************************************************************************
//    Add matrix a to b into c
// ***************************************************************************
void addMat( int a[N][N], int b[N][N], int c[N][N] )
{
	int i, j ;

#pragma omp parallel for
   	for( i = 0 ; i < N ; i++ )
#pragma omp simd
	for( j = 0 ; j < N ; j++ )
	    	c[i][j] = a[i][j] + b[i][j] ;

}


// ***************************************************************************
//    Compute the inner product of a and b
// ***************************************************************************
int innerProdMat( int a[N][N], int b[N][N] )
{
	int i, j, sum = 0 ;

#pragma omp parallel for reduction(+:sum)
	for( i = 0 ; i < N ; i++ ) 
#pragma omp simd reduction(+:sum)
	    for( j = 0 ; j < N ; j++ )
		    sum += a[i][j]*b[i][j] ;
	return sum ;
}


// ***************************************************************************
//    Compute in c the product of a and b
// ***************************************************************************
void mulMat( int a[N][N], int b[N][N], int c[N][N] )
{

#pragma omp parallel for collapse(2)
   	for( int i = 0 ; i < N ; i++ )
		for( int j = 0 ; j < N ; j++ ) {
		int q = 0.0 ;
#pragma omp simd reduction(+:q)
	        for( int k = 0 ; k < N ; k++ )
		        q += a[i][k] * b[k][j] ;
		c[i][j]=q ;
		}
}

// **************************************************************
int main(void)
{
int res ;

	setValMat( A ) ;
	setValMat( B ) ;
	addMat( A, B, C ) ;
	mulMat( C, B, A ) ;
	res = innerProdMat( A, B ) ;
	printf( "innerProdMat  %d\n", res ) ;
	return 0 ;
}
